/*
solid-genotype: call genotype on high-coverage SOLiD/BFAST re-sequencing data
Author: Jin Yu in Fuli's lab at HGSC of BCM
Date: 2011-02-11
*/

#include <stdio.h>
#include <string.h>
#include "sam.h"
#include "faidx.h"


typedef struct {
	int beg, end;
	samfile_t *in;
        faidx_t *fai;
        int ref_tid;
        char *ref;
} tmpstruct_t;

// callback for bam_fetch()
static int fetch_func(const bam1_t *b, void *data)
{
	bam_plbuf_t *buf = (bam_plbuf_t*)data;
	bam_plbuf_push(b, buf);
	return 0;
}
// callback for bam_plbuf_init()

/* qpos is 0-base */
static int evaluation(const bam_pileup1_t read)
{
   int read_length, qpos, dist3, cm, nqs, qual;
   float logit;
   read_length = read.b->core.l_qseq;
   qpos = read.qpos;
//calculate dist3'
   if (bam1_strand(read.b) == 0)  dist3 = read_length - qpos - 1;  else dist3 = qpos;
//calculate CM
   cm = bam_aux2i(bam_aux_get(read.b,"CM"));
//calculate base qual
   qual = bam1_qual(read.b)[read.qpos];
//first judgement
   if (qual < 18 || dist3 == 0) return 0;
//calculate NQS
   if (read_length <= 10) nqs = 0;
   else if (dist3 > read_length - 6) nqs = 1;
   else if (qpos >= 5 && read_length - qpos >= 6 && bam1_qual(read.b)[read.qpos] >= 20 && bam1_qual(read.b)[read.qpos - 1] >= 15 && bam1_qual(read.b)[read.qpos - 2] >= 15 && bam1_qual(read.b)[read.qpos - 3] >= 15 && bam1_qual(read.b)[read.qpos - 4] >= 15 && bam1_qual(read.b)[read.qpos - 5] >= 15 && bam1_qual(read.b)[read.qpos + 1] >= 15 && bam1_qual(read.b)[read.qpos + 2] >= 15 && bam1_qual(read.b)[read.qpos + 3] >= 15 && bam1_qual(read.b)[read.qpos + 4] >= 15 && bam1_qual(read.b)[read.qpos + 5] >= 15) nqs = 1;
   else nqs = 0;
//calculate logit regression value
   logit = 3.5 + 0.234 * cm + (-0.0845) * qual + (-0.00717) * dist3 + (-1.11) * nqs;
//second jusgement
   if (logit > 0) return 0;
   else return 1;
}

static int pileup_func(uint32_t tid, uint32_t pos, int n, const bam_pileup1_t *pl, void *data)
{
	tmpstruct_t *tmp = (tmpstruct_t*)data;
        // update reference seq when necessary
        if (tmp->fai && tid != tmp->ref_tid) {
            free(tmp->ref);
            tmp->ref = faidx_fetch_seq(tmp->fai, tmp->in->header->target_name[tid], 0, 0x7fffffff, (int*)tmp->in->header->target_len);
            tmp->ref_tid = tid;
        }

        int i, count_A, count_a, count_T, count_t, count_C, count_c, count_G, count_g, count_N, count_total;
        count_A = count_a = count_T = count_t = count_C = count_c = count_G = count_g = count_N = 0;
	if ((int)pos >= tmp->beg && (int)pos < tmp->end && n<1024)
            
            for (i=0; i<n; i++)
            {
                if (pl[i].b->core.qual < 255) continue;
                 if (bam_aux2i(bam_aux_get(pl[i].b,"NM")) * 10 >= pl[i].b->core.l_qseq) continue;

                switch (bam_nt16_rev_table[bam1_seqi(bam1_seq(pl[i].b), pl[i].qpos)]){
                    case 'A':
                        if (evaluation(pl[i]) == 1) count_A += 1;
                        else count_a += 1;
                        break;
                    case 'C':
                        if (evaluation(pl[i]) == 1) count_C += 1;
                        else count_c += 1;
                        break;
                    case 'G':
                        if (evaluation(pl[i]) == 1) count_G += 1;
                        else count_g += 1;
                        break;
                    case 'T':
                        if (evaluation(pl[i]) == 1) count_T += 1;
                        else count_t += 1;
                        break;
                    case 'N':
                        count_N += 1;
                        break;
                }

/*                if (evaluation(pl[i]) == 1) printf("%c",bam_nt16_rev_table[bam1_seqi(bam1_seq(pl[i].b), pl[i].qpos)]);
                else printf("%c",bam_nt16_rev_table[bam1_seqi(bam1_seq(pl[i].b), pl[i].qpos)] + 32);*/

            }
        
        count_total = count_A + count_T + count_C + count_G;
        char genotype[5]="";
        if (10 * count_A >= count_total && count_A >= 2) strcat(genotype,"A");
        if (10 * count_T >= count_total && count_T >= 2) strcat(genotype,"T");
        if (10 * count_C >= count_total && count_C >= 2) strcat(genotype,"C");
        if (10 * count_G >= count_total && count_G >= 2) strcat(genotype,"G");

        if (count_total >= 8)
        {
            if (strlen(genotype)==1) strcat(genotype,genotype);
            printf("%s\t%d\t%c\t%s\t", tmp->in->header->target_name[tid], pos + 1, tmp->ref[pos], genotype);
            printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t", count_A, count_C, count_G, count_T, count_N, count_a, count_c, count_g, count_t, count_total, n);
            printf("\n");
        }

//            printf("%s\t%d\t%c\t%d\t%d\t%d\t%d\t%d\n", tmp->in->header->target_name[tid], pos + 1, tmp->ref[pos], n, pl[0].qpos, pl[n-1].qpos, pl[0].b->core.flag, evaluation(pl[0]));
	return 0;
}

int main(int argc, char *argv[])
{
	tmpstruct_t tmp;
	if (argc < 3) {
		fprintf(stderr, "Usage: solid-genotyper <in.bam> <ref.fa> [by chromosome]\n");
		return 1;
	}
	tmp.beg = 0; tmp.end = 0x7fffffff;
	tmp.in = samopen(argv[1], "rb", 0);
        tmp.fai = fai_load(argv[2]);
        tmp.ref = faidx_fetch_seq(tmp.fai, tmp.in->header->target_name[0], 0, 0x7fffffff, (int*)tmp.in->header->target_len);
	if (tmp.in == 0) {
		fprintf(stderr, "Fail to open BAM file %s\n", argv[1]);
		return 1;
	}
	if (argc == 3) { // if a region is not specified
		sampileup(tmp.in, -1, pileup_func, &tmp);
	} else {//fixme: only work by given chromosome, not work by given region
		int ref;
		bam_index_t *idx;
		bam_plbuf_t *buf;
		idx = bam_index_load(argv[1]); // load BAM index
		if (idx == 0) {
			fprintf(stderr, "BAM indexing file is not available.\n");
			return 1;
		}
		bam_parse_region(tmp.in->header, argv[3], &ref, &tmp.beg, &tmp.end); // parse the region
		if (ref < 0) {
			fprintf(stderr, "Invalid region %s\n", argv[3]);
			return 1;
		}
		buf = bam_plbuf_init(pileup_func, &tmp); // initialize pileup
		bam_fetch(tmp.in->x.bam, idx, ref, tmp.beg, tmp.end, buf, fetch_func);
		bam_plbuf_push(0, buf); // finalize pileup
		bam_index_destroy(idx);
		bam_plbuf_destroy(buf);
	}
	samclose(tmp.in);
        free(tmp.ref);
        free(tmp.fai);
	return 0;
}
